/*-----------------------------------------------------------------------------
  Project:  QRE
  File:     WBMHandler.h

  Copyright (C)2010 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Revision: 1.6 $
 *----------------------------------------------------------------------------*/
#ifndef WBMHANDLER_H_
# define WBMHANDLER_H_

/// @file
//
///
/// See <a href="http://en.wikipedia.org/wiki/Windows_bitmap">
/// http://en.wikipedia.org/wiki/Windows_bitmap</a>
//
/// @version 1.0

# include <nn/fs.h>

typedef nn::fs::FileInputStream ISTREAM;
typedef nn::fs::FileOutputStream OSTREAM;

// Size of each bitmap header
const uint32_t BITMAPFILEHEADER_SIZE = 14;    //
const uint32_t BHS_OS2_V1_SIZE = 12;        //
const uint32_t BHS_OS2_V2_SIZE = 64;        //
const uint32_t BHS_WIN_V3_SIZE = 40;        //
const uint32_t BHS_WIN_V4_SIZE = 108;        //
const uint32_t BHS_WIN_V5_SIZE = 124;        //

/* Please see man pages for details 
 
 
 
 
 
 
 
 
 */
template<typename T>
inline ISTREAM &ReadFile(ISTREAM &is, T *ptr, size_t read_size = 0)
{
    if (!read_size)
        read_size = sizeof(T);
    is.Read(reinterpret_cast<char *>(ptr), read_size);
    return is;
}

/* Please see man pages for details 
 
 
 
 
 
 
 
 
 */
template<typename T>
inline OSTREAM &WriteFile(OSTREAM &os, T *ptr, size_t write_size = 0)
{
    if (!write_size)
        write_size = sizeof(T);
    os.Write(reinterpret_cast<char *>(ptr), write_size, false);
    return os;
}

/* Please see man pages for details 
 
 
 
 */
template<typename T>
inline T round4(T n)
{
    return (n + (static_cast<T>(3))) & ~(static_cast<T>(3));
}

//
///
/// Size is 12 bytes
class BitMapCoreHeader {
protected:
    uint32_t header_size;    //
    uint32_t width;            //
    int32_t height;            //
    uint16_t plane;            //
    uint16_t color_bits;    //

public:
    //
    //
    ///          (Interface parameters for derived classes, generally must use default parameters)
    BitMapCoreHeader(const uint32_t l = BHS_OS2_V1_SIZE):
        header_size(l), width(0), height(0), plane(1), color_bits(24) {}
    //
    ///
    /// Load header information from specified input stream.
    //
    //
    ///          (Interface parameters for derived classes, generally must use default parameters)
    BitMapCoreHeader(ISTREAM &is,
                     const uint32_t l = BHS_OS2_V1_SIZE): header_size(l) {
        read_header(is);
    }
    /// Destructor
    virtual ~BitMapCoreHeader() {}

    virtual int read_header(ISTREAM &);
    virtual int write_header(OSTREAM &);

    //
    //
    uint32_t get_width() { return width; }
    //
    //
    int32_t get_height() { return height; }
    //
    //
    uint32_t get_bits() { return color_bits; }
    //
    //
    virtual uint32_t get_horizontal_PPM() { return 0; }
    //
    //
    virtual uint32_t get_vertical_PPM() { return 0; }
    //
    //
    virtual uint32_t get_image_size() {
        return height * round4(width * color_bits / 8) * (height > 0 ? 1 : -1);
    }
    //
    //
    //
    virtual uint32_t set_width(const uint32_t w) {
        width = w; return width;
    }
    //
    //
    //
    virtual int32_t set_height(const int32_t h) {
        height = h; return height;
    }
    //
    //
    virtual uint32_t set_horizontal_PPM(const uint32_t) { return 0; }
    //
    //
    virtual uint32_t set_vertical_PPM(const uint32_t) { return 0; }

protected:
    virtual int read_sizeinfo(ISTREAM &);
    virtual int write_sizeinfo(OSTREAM &);
};

//
///
/// Size is 40 bytes
class BitMapInfoHeader: public BitMapCoreHeader {
protected:
    uint32_t compress;            //
    uint32_t image_size;        //
    uint32_t horizontal_PPM;    //
    uint32_t vertical_PPM;        //
    uint32_t num_of_color;        //
    uint32_t important_color;    //

public:
    //
    //
    ///          (Interface parameters for derived classes, generally must use default parameters)
    BitMapInfoHeader(const uint32_t l = BHS_WIN_V3_SIZE):
        BitMapCoreHeader(l), compress(0), image_size(0),
        horizontal_PPM(2834), vertical_PPM(2834),
        num_of_color(0), important_color(0) {}
    //
    ///
    /// Load header information from specified input stream.
    //
    //
    ///          (Interface parameters for derived classes, generally must use default parameters)
    BitMapInfoHeader(ISTREAM &is, const uint32_t l = BHS_WIN_V3_SIZE):
        BitMapCoreHeader(is, l) {

        read_header(is);
    }
    /// Destructor
    virtual ~BitMapInfoHeader() {}

    virtual int read_header(ISTREAM &);
    virtual int write_header(OSTREAM &);

    //
    //
    virtual uint32_t get_horizontal_PPM() { return horizontal_PPM; }
    //
    //
    virtual uint32_t get_vertical_PPM() { return vertical_PPM; }
    //
    //
    virtual uint32_t get_image_size() { return image_size; }
    //
    //
    //
    virtual uint32_t set_horizontal_PPM(const uint32_t p) {
        horizontal_PPM = p;
        return horizontal_PPM;
    }
    //
    //
    //
    virtual uint32_t set_vertical_PPM(const uint32_t p) {
        vertical_PPM = p;
        return vertical_PPM;
    }

    //
    //
    //
    virtual uint32_t set_width(const uint32_t w) {
        BitMapCoreHeader::set_width(w);
        recalc_size();
        return width;
    }
    //
    //
    //
    virtual int32_t set_height(const int32_t h) {
        BitMapCoreHeader::set_height(h);
        recalc_size();
        return height;
    }

protected:
    virtual int read_sizeinfo(ISTREAM &);
    virtual int write_sizeinfo(OSTREAM &);

private:
    //
    ///
    /// Recalculate file size based on saved size information.
    //
    //
    int recalc_size()
    {
        image_size = BitMapCoreHeader::get_image_size();
        return 0;
    }
};

//
///
/// +----------------------+-----------------------+------------------------+--<br>
/// |2-byte identifier "BM"|4-byte file size (byte)|2-byte reserved region 1|...<br>
/// +----------------------+-----------------------+------------------------+--<br>
/// <br>
///  --+------------------------+-------------------------------------------+<br>
/// ...|2-byte reserved region 2|4-byte bitmap data starting position (byte)|<br>
///  --+------------------------+-------------------------------------------+<br>
class BitMapFileHeader {
    NN_PADDING2;
    char magic[2];                    //
    uint32_t size;                    //
    uint16_t reserve_1;                //
    uint16_t reserve_2;                //
    uint32_t offset;                //
    BitMapCoreHeader *core_header;    //
    uint8_t *color_map;                //
    uint32_t color_map_size;        //
public:
    //
    ///
    /// Initialize the pointer region.
    /// Start by initializing the generic header.
    BitMapFileHeader():
        size(BITMAPFILEHEADER_SIZE + BHS_WIN_V3_SIZE),
        reserve_1(0), reserve_2(0),
        offset(BITMAPFILEHEADER_SIZE + BHS_WIN_V3_SIZE),
        core_header(NULL), color_map(NULL), color_map_size(0) {
        magic[0] = 'B';
        magic[1] = 'M';
        core_header 
            = reinterpret_cast<BitMapCoreHeader *>(new BitMapInfoHeader());
    }
    //
    ///
    /// Load header information from specified input stream.
    //
    BitMapFileHeader(ISTREAM &is):
        core_header(NULL), color_map(NULL) {

        read_header(is);
    }
    //
    ///
    /// Deallocate assigned memory.
    virtual ~BitMapFileHeader() {
        if (core_header) delete core_header;
        if (color_map) delete [] color_map;
    }

    int read_header(ISTREAM &);
    int write_header(OSTREAM &);

    //
    //
    uint32_t get_width() {
        return (core_header) ? core_header->get_width(): 0;
    }
    //
    //
    int32_t get_height() {
        return (core_header) ? core_header->get_height(): 0;
    }
    //
    //
    uint32_t get_bits() {
        return (core_header) ? core_header->get_bits(): 0;
    }
    //
    //
    uint32_t get_horizontal_PPM() {
        return (core_header) ? core_header->get_horizontal_PPM(): 0;
    }
    //
    //
    uint32_t get_vertical_PPM() {
        return (core_header) ? core_header->get_vertical_PPM(): 0;
    }

    //
    //
    //
    //
    int set_width(const uint32_t w) {
        if (!core_header)
            return -1;
        core_header->set_width(w);
        return recalc_size();
    }
    //
    //
    //
    //
    int set_height(const int32_t h) {
        if (!core_header)
            return -1;
        core_header->set_height(h);
        return recalc_size();
    }

    //
    //
    //
    //
    uint32_t set_horizontal_PPM(const uint32_t p) {
        if (!core_header)
            return 0;
        return core_header->set_horizontal_PPM(p);
    }
    //
    //
    //
    //
    uint32_t set_vertical_PPM(const uint32_t p) {
        if (!core_header)
            return 0;
        return core_header->set_vertical_PPM(p);
    }

    uint32_t get_header_size() {
        return offset;
    }

    uint32_t get_image_size() {
        return size - offset;
    }

private:
    //
    ///
    /// Recalculate file size based on saved size information.
    //
    //
    int recalc_size() {
        // File size is bitmap data starting position plus image size
        // 
        size = static_cast<uint32_t>(offset + core_header->get_image_size());
        return 0;
    }
};

#endif // WBMHANDLER_H_
/* EOF */
